"""
Project: BlueMath_tk
Sub-Module: downloaders
Author: GeoOcean Research Group, Universidad de Cantabria
Creation Date: 9 August 2024
License: MIT
Repository: https://github.com/GeoOcean/BlueMath_tk.git
Status: Under development (Working)
"""

# Import essential functions/classes to be available at the package level.
from ._base_downloaders import BaseDownloader
from .copernicus.copernicus_downloader import CopernicusDownloader

# Optionally, define the module's `__all__` variable to control what gets imported when using `from module import *`.
__all__ = [
    "BaseDownloader",
    "CopernicusDownloader",
]



from .._base_wrappers import BaseModelWrapper


class XBeachModelWrapper(BaseModelWrapper):
    """
    Wrapper for the XBeach model.
    https://xbeach.readthedocs.io/en/latest/

    Attributes
    ----------
    xbeach_exec : str
        The XBeach executable path.
    default_parameters : dict
        The default parameters type for the model.

    Methods
    -------
    set_xbeach_exec(xbeach_exec: str) -> None
        Set the XBeach executable path.
    run_model() -> None
        Run the XBeach model for the specified case.
    """

    default_parameters = {
        "spectra": str,
    }

    def __init__(
        self,
        templates_dir: str,
        templates_name: dict,
        model_parameters: dict,
        output_dir: str,
    ) -> None:
        """
        Initialize the SWASH model wrapper.

        Parameters
        ----------
        templates_dir : str
            The directory where the templates are stored.
        templates_name : list
            The names of the templates.
        model_parameters : dict
            The parameters to be used in the templates.
        output_dir : str
            The directory where the output files will be saved.
        """

        super().__init__(
            templates_dir=templates_dir,
            templates_name=templates_name,
            model_parameters=model_parameters,
            output_dir=output_dir,
            default_parameters=self.default_parameters,
        )
        self.set_logger_name(self.__class__.__name__)
        self._xbeach_exec = None

    @property
    def xbeach_exec(self) -> str:
        return self._xbeach_exec

    def set_xbeach_exec(self, xbeach_exec: str) -> None:
        self._xbeach_exec = xbeach_exec

    def run_model(self) -> None:
        pass



#!/usr/bin/env python

"""
Usage: era5_download.py YEARS MONTHS [--var=<var>] [--base-path=<base-path>] [--token=<token>] [--debug] [--check] [--force]
era5_download.py (-h | --help | --version)
era5_download.py --list-vars
era5_download.py --list-datasets
era5_download.py --show-info

This script downloads ERA5 reanalysis data for multiple variables and saves them as NetCDF files.

Arguments:
  YEARS                     Years of the data to download, separated by commas
                              To download a single year, use the format YYYY
                              To download a range of years, use the format YYYY-YYYY
  MONTHS                    Months of the data to download, separated by commas
                              To download a single month, use the format MM
                              To download a range of months, use the format MM-MM

Options:
  --var=<var>               List of variable values separated by commas.
  --base-path=<base-path>   Base path to save the data [default: /lustre/geocean/DATA/Copernicus/].
  --token=<token>           Token for the Climate Data Store API.
  --debug                   Debug mode (it is recommended to use this mode).
  --check                   If just file checking is required.
  --force                   Overwrite the data in case it exists.
  --list-vars               Return a list with the available variables.
  --list-datasets           Return a list with the available datasets.
  --show-info               Show information about the variables.

Examples:
    era5_download.py 2014,2016 1,10 --var tp,sst --debug
    era5_download.py 2014,2016 1,10 --var tp,sst --debug --check
    era5_download.py --list-vars
    era5_download.py --list-datasets
    era5_download.py --show-info
"""

from docopt import docopt
from bluemath_tk.downloaders.copernicus.copernicus_downloader import (
    CopernicusDownloader,
)

# Parse the command line arguments
args = docopt(__doc__)
print(args)

# Create the CopernicusDownloader object
copernicus_downloader = CopernicusDownloader(
    product="ERA5",
    token=args["--token"],
    base_path_to_download=args["--base-path"] or "/lustre/geocean/DATA/Copernicus/",
    debug=args["--debug"],
    check=args["--check"],
)

if args["--list-vars"]:
    print(copernicus_downloader.list_variables())
    exit()
if args["--list-datasets"]:
    print(copernicus_downloader.list_datasets())
    exit()
if args["--show-info"]:
    print(copernicus_downloader.show_markdown_table())
    exit()

years = (
    args["YEARS"].split(",")
    if "-" not in args["YEARS"]
    else list(
        range(int(args["YEARS"].split("-")[0]), int(args["YEARS"].split("-")[1]) + 1)
    )
)
months = (
    args["MONTHS"].split(",")
    if "-" not in args["MONTHS"]
    else list(
        range(int(args["MONTHS"].split("-")[0]), int(args["MONTHS"].split("-")[1]) + 1)
    )
)
# Ocean variables are used by default, as right now, DestinE program is used for atmosphere variables
variables = (
    args["--var"].split(",")
    if args["--var"]
    else copernicus_downloader.list_variables(type="ocean")
)

output = copernicus_downloader.download_data(
    variables=variables,
    years=years,
    months=months,
    force=args["--force"],
)

print(output)



from abc import ABC, abstractmethod
from typing import Union
import matplotlib.pyplot as plt
from matplotlib.colors import Colormap
import plotly.graph_objects as go
import cartopy.crs as ccrs


class BasePlotting(ABC):
    """
    Abstract base class for handling default plotting functionalities across the project.
    """

    def __init__(self):
        self.default_line_color = "blue"
        self.default_line_style = "-"
        self.default_scatter_color = "red"
        self.default_scatter_size = 10
        self.default_marker = "o"

    @abstractmethod
    def plot_line(self, x, y):
        """
        Abstract method for plotting a line.
        Should be implemented by subclasses.
        """
        pass

    @abstractmethod
    def plot_scatter(self, x, y):
        """
        Abstract method for plotting a scatter plot.
        Should be implemented by subclasses.
        """
        pass

    @abstractmethod
    def plot_map(self, markers=None):
        """
        Abstract method for plotting a map.
        Should be implemented by subclasses.
        """
        pass

    def get_list_of_colors_for_colormap(
        self, cmap: Union[str, Colormap], num_colors: int
    ) -> list:
        """
        Get a list of colors from a colormap.

        Parameters
        ----------
        cmap : str or Colormap
            The colormap to use.
        num_colors : int
            The number of colors to generate.

        Returns
        -------
        list
            A list of colors generated from the colormap.
        """

        if isinstance(cmap, str):
            cmap = plt.get_cmap(cmap)
        return [cmap(i) for i in range(0, 256, 256 // num_colors)]


class DefaultStaticPlotting(BasePlotting):
    """
    Concrete implementation of BasePlotting with static plotting behaviors.
    """

    # Class-level dictionary for default settings
    templates = {
        "default": {
            "line_color": "blue",
            "line_style": "-",
            "scatter_color": "red",
            "scatter_size": 10,
            "marker": "o",
        }
    }

    def __init__(self, **kwargs):
        """
        Initialize an instance of the DefaultStaticPlotting class.

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments to override default settings.
            Valid keys are:
            - template: str, name of the template to use, defaults to "default"
            - line_color: str, color of the line, defaults to blue
            - line_style: str, style of the line, defaults to solid line
            - scatter_color: str, color of the scatter plot, defaults to red
            - scatter_size: int, size of the scatter plot, defaults to 10
            - marker: str, marker of the scatter plot, defaults to circle

        Returns
        -------
        None

        Notes
        -----
        - If no keyword arguments are provided, the default template is used.
        - If a keyword argument is provided, it will override the corresponding default setting.
        - Any other provided keyword arguments will be set as instance attributes.
        """
        super().__init__()
        # Update instance attributes with either default template or passed-in values / template
        for key, value in self.templates.get(
            kwargs.get("template", "default"), self.templates.get("default")
        ).items():
            setattr(self, key, kwargs.get(key, value))
        for key, value in kwargs.items():
            setattr(self, key, value)

    def get_subplots(self, **kwargs):
        fig, ax = plt.subplots(**kwargs)
        return fig, ax

    def plot_line(self, ax, **kwargs):
        ax.plot(**kwargs)
        self.set_grid(ax)

    def plot_scatter(self, ax, **kwargs):
        ax.scatter(**kwargs)
        self.set_grid(ax)

    def plot_pie(self, ax, **kwargs):
        ax.pie(**kwargs)

    def plot_map(self, ax, **kwargs):
        ax.set_global()
        ax.coastlines()

    def set_title(self, ax, title="Plot Title"):
        """
        Sets the title for a given axis.
        """
        ax.set_title(title)

    def set_xlim(self, ax, xmin, xmax):
        """
        Sets the x-axis limits for a given axis.
        """
        ax.set_xlim(xmin, xmax)

    def set_ylim(self, ax, ymin, ymax):
        """
        Sets the y-axis limits for a given axis.
        """
        ax.set_ylim(ymin, ymax)

    def set_xlabel(self, ax, xlabel="X-axis"):
        """
        Sets the x-axis label for a given axis.
        """
        ax.set_xlabel(xlabel)

    def set_ylabel(self, ax, ylabel="Y-axis"):
        """
        Sets the y-axis label for a given axis.
        """
        ax.set_ylabel(ylabel)

    def set_grid(self, ax, grid=True):
        """
        Sets the grid for a given axis.
        """
        ax.grid(grid)


if __name__ == "__main__":
    plotting = DefaultStaticPlotting()
    plotting.get_subplots(figsize=(10, 6))


class DefaultInteractivePlotting(BasePlotting):
    """
    Concrete implementation of BasePlotting with interactive plotting behaviors.
    """

    def __init__(self):
        super().__init__()

    def plot_line(self, x, y):
        fig = go.Figure()
        fig.add_trace(
            go.Scatter(x=x, y=y, mode="lines", line=dict(color=self.default_line_color))
        )
        fig.update_layout(
            title="Interactive Line Plot", xaxis_title="X-axis", yaxis_title="Y-axis"
        )
        fig.show()

    def plot_scatter(self, x, y):
        fig = go.Figure()
        fig.add_trace(
            go.Scatter(
                x=x, y=y, mode="markers", marker=dict(color=self.default_scatter_color)
            )
        )
        fig.update_layout(
            title="Interactive Scatter Plot", xaxis_title="X-axis", yaxis_title="Y-axis"
        )
        fig.show()

    def plot_map(self, markers=None):
        fig = go.Figure(
            go.Scattermapbox(
                lat=[marker[0] for marker in markers] if markers else [],
                lon=[marker[1] for marker in markers] if markers else [],
                mode="markers",
                marker=go.scattermapbox.Marker(size=10, color="red"),
            )
        )
        fig.update_layout(
            mapbox=dict(
                style="open-street-map",
                center=dict(
                    lat=self.default_map_center[0], lon=self.default_map_center[1]
                ),
                zoom=self.default_map_zoom_start,
            ),
            title="Interactive Map with Plotly",
        )
        fig.show()



import os
import json
from typing import List
import calendar
import cdsapi
import xarray as xr
from .._base_downloaders import BaseDownloader

config = {
    "url": "https://cds.climate.copernicus.eu/api",  # /v2?
    "key": "your-api-token",
}


class CopernicusDownloader(BaseDownloader):
    """
    This is the main class to download data from the Copernicus Climate Data Store.

    Attributes
    ----------
    product : str
        The product to download data from. Currently only ERA5 is supported.
    product_config : dict
        The configuration for the product to download data from.
    client : cdsapi.Client
        The client to interact with the Copernicus Climate Data Store API.

    Notes
    -----
    - The configuration for the products is stored in the `products_configs` attribute.
    - The configuration for the ERA5 product is stored in the `ERA5` folder.
    """

    products_configs = {
        "ERA5": json.load(
            open(os.path.join(os.path.dirname(__file__), "ERA5", "ERA5_config.json"))
            # open(
            #     "/home/grupos/geocean/tausiaj/BlueMath_tk/bluemath_tk/downloaders/copernicus/ERA5/ERA5_config.json"
            # )
        )
    }

    def __init__(
        self,
        product: str,
        base_path_to_download: str,
        token: str = None,
        debug: bool = True,
        check: bool = False,
    ) -> None:
        """
        This is the constructor for the CopernicusDownloader class.

        Parameters
        ----------
        product : str
            The product to download data from. Currently only ERA5 is supported.
        base_path_to_download : str
            The base path to download the data to.
        token : str, optional
            The API token to use to download data. Defaults to None.
        debug : bool, optional
            Whether to run in debug mode. Defaults to True.
        check : bool, optional
            Whether to just check the data. Defaults to False.

        Raises
        ------
        ValueError
            If the product configuration is not found.
        """

        super().__init__(
            base_path_to_download=base_path_to_download, debug=debug, check=check
        )
        self._product = product
        self._product_config = self.products_configs.get(product)
        if self._product_config is None:
            raise ValueError(f"{product} configuration not found")
        self.set_logger_name(f"CopernicusDownloader - {product}")
        if not self.check:
            self._client = cdsapi.Client(
                url=config["url"], key=token or config["key"], debug=self.debug
            )
            self.logger.info("---- DOWNLOADING DATA ----")
        else:
            self.logger.info("---- CHECKING DATA ----")

    @property
    def product(self) -> str:
        return self._product

    @property
    def product_config(self) -> dict:
        return self._product_config

    @property
    def client(self) -> cdsapi.Client:
        return self._client

    def list_variables(self, type: str = None) -> List[str]:
        """
        Lists the variables available for the product.
        Filtering by type if provided.

        Parameters
        ----------
        type : str, optional
            The type of variables to list. Default is None.

        Returns
        -------
        List[str]
            The list of variables available for the product.
        """

        if type == "ocean":
            return [
                var_name
                for var_name, var_info in self.product_config["variables"].items()
                if var_info["type"] == "ocean"
            ]
        return list(self.product_config["variables"].keys())

    def list_datasets(self) -> List[str]:
        """
        Lists the datasets available for the product.

        Returns
        -------
        List[str]
            The list of datasets available for the product.
        """

        return list(self.product_config["datasets"].keys())

    def show_markdown_table(self) -> None:
        """
        Create a Markdown table from the configuration dictionary and print it.
        """

        # Define the table headers
        headers = ["name", "long_name", "units", "type"]
        header_line = "| " + " | ".join(headers) + " |"
        separator_line = (
            "| " + " | ".join(["-" * len(header) for header in headers]) + " |"
        )

        # Initialize the table with headers
        table_lines = [header_line, separator_line]

        # Add rows for each variable
        for var_name, var_info in self.product_config["variables"].items():
            long_name = var_info.get("long_name", "")
            units = var_info.get("units", "")
            type = var_info.get("type", "")
            row = f"| {var_name} | {long_name} | {units} | {type} |"
            table_lines.append(row)

        # Print the table
        print("\n".join(table_lines))

    def download_data(self, *args, **kwargs) -> str:
        """
        Downloads the data for the product.

        Parameters
        ----------
        *args
            The arguments to pass to the download function.
        **kwargs
            The keyword arguments to pass to the download function.

        Returns
        -------
        str
            The message with the fully downloaded files and the not fully downloaded files.

        Raises
        ------
        ValueError
            If the product is not supported.
        """

        if self.product == "ERA5":
            return self.download_data_era5(*args, **kwargs)
        else:
            raise ValueError(f"Download for product {self.product} not supported")

    def download_data_era5(
        self,
        variables: List[str],
        years: List[str],
        months: List[str],
        day: List[str] = None,
        time: List[str] = None,
        product_type: str = "reanalysis",
        data_format: str = "netcdf_legacy",
        download_format: str = "unarchived",
        force: bool = False,
    ) -> str:
        """
        Downloads the data for the ERA5 product.

        Parameters
        ----------
        variables : List[str]
            The list of variables to download.
        years : List[str]
            The list of years to download.
        months : List[str]
            The list of months to download.
        day : List[str], optional
            The list of days to download. Defaults to None.
        time : List[str], optional
            The list of times to download. Defaults to None.
        product_type : str, optional
            The product type to download. Defaults to "reanalysis".
        data_format : str, optional
            The data format to download. Defaults to "netcdf_legacy".
            This format is maintained to legacy to allow Thredds server to read the files.
        download_format : str, optional
            The download format to use. Defaults to "unarchived".
        force : bool, optional
            Whether to force the download. Defaults to False.

        Returns
        -------
        str
            The message with the fully downloaded files and the not fully downloaded files.
            Error files are also included.

        Raises
        ------
        ValueError
            If the variables, years, months, day, time, product_type, data_format, or download_format are not valid
            or if the force is not a boolean.

        Notes
        -----
        - The variables, years, months, day, and time must be lists of strings.
        - The product_type, data_format, and download_format must be strings.
        - The force must be a boolean.
        - Leave variables empty to download all variables.
        - Leave day empty to download all days.
        - Leave time empty to download all times.
        """

        if not isinstance(variables, list):
            raise ValueError("Variables must be a list of strings")
        elif len(variables) == 0:
            variables = list(self.product_config["variables"].keys())
            self.logger.info(f"Variables not provided. Using {variables}")
        if not isinstance(years, list) or len(years) == 0:
            raise ValueError("Years must be a non-empty list of strings")
        if not isinstance(months, list) or len(months) == 0:
            raise ValueError("Months must be a non-empty list of strings")
        if day is not None:
            if not isinstance(day, list) or len(day) == 0:
                raise ValueError("Day must be a non-empty list of strings")
        else:
            day = [f"{day:02d}" for day in range(1, 32)]
            self.logger.info(f"Day not provided. Using {day}")
        if time is not None:
            if not isinstance(time, list) or len(time) == 0:
                raise ValueError("Time must be a non-empty list of strings")
        else:
            time = [f"{hour:02d}:00" for hour in range(24)]
            self.logger.info(f"Time not provided. Using {time}")
        if not isinstance(product_type, str):
            raise ValueError("Product type must be a string")
        if not isinstance(data_format, str):
            raise ValueError("Data format must be a string")
        if not isinstance(download_format, str):
            raise ValueError("Download format must be a string")
        if not isinstance(force, bool):
            raise ValueError("Force must be a boolean")

        fully_downloaded_files: List[str] = []
        NOT_fullly_downloaded_files: List[str] = []
        error_files: List[str] = []

        for variable in variables:
            for year in years:
                year = f"{int(year):04d}"  # Ensure year is 4 digits
                for month in months:
                    month = f"{int(month):02d}"  # Ensure month is 2 digits
                    variable_config = self.product_config["variables"].get(variable)
                    if variable_config is None:
                        self.logger.error(
                            f"Variable {variable} not found in product configuration file"
                        )
                        continue
                    variable_dataset = self.product_config["datasets"].get(
                        variable_config["dataset"]
                    )
                    if variable_dataset is None:
                        self.logger.error(
                            f"Dataset {variable_config['dataset']} not found in product configuration file"
                        )
                        continue

                    template_for_variable = variable_dataset["template"].copy()
                    template_for_variable["variable"] = variable_config["cds_name"]
                    template_for_variable["year"] = year
                    template_for_variable["month"] = month
                    template_for_variable["day"] = day
                    template_for_variable["time"] = time
                    template_for_variable["product_type"] = product_type
                    template_for_variable["data_format"] = data_format
                    template_for_variable["download_format"] = download_format

                    skip_because_of_manadatory_fields = False
                    for mandatory_field in variable_dataset["mandatory_fields"]:
                        try:
                            if template_for_variable.get(mandatory_field) is None:
                                template_for_variable[mandatory_field] = (
                                    variable_config[mandatory_field]
                                )
                        except KeyError:
                            self.logger.error(
                                f"Mandotory field {mandatory_field} not found in variable configuration file for {variable}"
                            )
                            skip_because_of_manadatory_fields = True
                    if skip_because_of_manadatory_fields:
                        continue

                    # Create the output file name once request is properly formatted
                    output_nc_file = os.path.join(
                        self.base_path_to_download,
                        self.product,
                        variable_config["dataset"],
                        variable_config["type"],
                        product_type,
                        variable_config["nc_name"],
                        year,
                        f"{variable_config['nc_name']}_{year}{month}.nc",
                    )
                    # Create the output directory if it does not exist
                    os.makedirs(os.path.dirname(output_nc_file), exist_ok=True)

                    self.logger.info(f"""
                                     
                        Analyzing {output_nc_file}
                                     
                    """)

                    try:
                        if self.check or not force:
                            if os.path.exists(output_nc_file):
                                self.logger.debug(
                                    f"Checking {output_nc_file} file is complete"
                                )
                                try:
                                    nc = xr.open_dataset(output_nc_file)
                                    _, last_day = calendar.monthrange(
                                        int(year), int(month)
                                    )
                                    last_hour = f"{year}-{int(month):02d}-{last_day}T23"
                                    last_hour_nc = str(nc.time[-1].values)
                                    nc.close()
                                    if last_hour not in last_hour_nc:
                                        self.logger.debug(
                                            f"{output_nc_file} ends at {last_hour_nc} instead of {last_hour}"
                                        )
                                        if self.check:
                                            NOT_fullly_downloaded_files.append(
                                                output_nc_file
                                            )
                                        else:
                                            self.logger.debug(
                                                f"Downloading: {variable} to {output_nc_file} because it is not complete"
                                            )
                                            self.client.retrieve(
                                                name=variable_config["dataset"],
                                                request=template_for_variable,
                                                target=output_nc_file,
                                            )
                                            fully_downloaded_files.append(
                                                output_nc_file
                                            )
                                    else:
                                        fully_downloaded_files.append(output_nc_file)
                                except Exception as e:
                                    self.logger.error(
                                        f"Error was raised opening {output_nc_file} - {e}, re-downloading..."
                                    )
                                    if self.check:
                                        NOT_fullly_downloaded_files.append(
                                            output_nc_file
                                        )
                                    else:
                                        self.logger.debug(
                                            f"Downloading: {variable} to {output_nc_file} because it is not complete"
                                        )
                                        self.client.retrieve(
                                            name=variable_config["dataset"],
                                            request=template_for_variable,
                                            target=output_nc_file,
                                        )
                                        fully_downloaded_files.append(output_nc_file)
                            elif self.check:
                                NOT_fullly_downloaded_files.append(output_nc_file)
                            else:
                                self.logger.debug(
                                    f"Downloading: {variable} to {output_nc_file}"
                                )
                                self.client.retrieve(
                                    name=variable_config["dataset"],
                                    request=template_for_variable,
                                    target=output_nc_file,
                                )
                                fully_downloaded_files.append(output_nc_file)
                        else:
                            self.logger.debug(
                                f"Downloading: {variable} to {output_nc_file}"
                            )
                            self.client.retrieve(
                                name=variable_config["dataset"],
                                request=template_for_variable,
                                target=output_nc_file,
                            )
                            fully_downloaded_files.append(output_nc_file)

                    except Exception as e:
                        self.logger.error(f"""
                                          
                            Skippping {output_nc_file} for {e}

                        """)
                        error_files.append(output_nc_file)

        fully_downloaded_files_str = "\n".join(fully_downloaded_files)
        NOT_fullly_downloaded_files_str = "\n".join(NOT_fullly_downloaded_files)
        error_files = "\n".join(error_files)

        return f"""
            Fully downloaded files:
            {fully_downloaded_files_str}
            Not fully downloaded files:
            {NOT_fullly_downloaded_files_str}
            Error files:
            {error_files}
        """


if __name__ == "__main__":
    copernicus_downloader = CopernicusDownloader(
        product="ERA5",
        base_path_to_download="/home/tausiaj/DATA/Copernicus/",
        debug=True,
        check=False,
    )
    copernicus_downloader.download_data(
        variables=["geo500", "tp", "p140122"], years=["2021"], months=["01"]
    )



import unittest
import numpy as np
import xarray as xr
from bluemath_tk.datamining.pca import PCA


def get_2d_dataset():
    # Define the coordinates
    coord1 = np.linspace(-100, 100, 20)
    coord2 = np.linspace(-100, 100, 20)
    coord3 = np.arange(1, 50)

    # Create a meshgrid
    coord1, coord2, coord3 = np.meshgrid(coord1, coord2, coord3, indexing="ij")

    # Create a 3D dataset
    X = (
        np.sin(np.radians(coord1)) * np.cos(np.radians(coord2)) * np.sin(coord3)
        + np.sin(2 * np.radians(coord1))
        * np.cos(2 * np.radians(coord2))
        * np.sin(2 * coord3)
        + np.sin(3 * np.radians(coord1))
        * np.cos(3 * np.radians(coord2))
        * np.sin(3 * coord3)
    )
    # Create a 3D dataset
    Y = -np.sin(X)

    # Create an xarray dataset
    ds = xr.Dataset(
        {
            "X": (["coord1", "coord2", "coord3"], X),
            "Y": (["coord1", "coord2", "coord3"], Y),
        },
        coords={
            "coord1": coord1[:, 0, 0],
            "coord2": coord2[0, :, 0],
            "coord3": coord3[0, 0, :],
        },
    )

    return ds


class TestPCA(unittest.TestCase):
    def setUp(self):
        self.ds = get_2d_dataset()
        self.pca = PCA(n_components=5)
        self.ipca = PCA(n_components=5, is_incremental=True)

    def test_fit(self):
        self.pca.fit(
            data=self.ds,
            vars_to_stack=["X", "Y"],
            coords_to_stack=["coord1", "coord2"],
            pca_dim_for_rows="coord3",
        )
        self.assertEqual(self.pca.is_fitted, True)

    def test_transform(self):
        self.pca.fit(
            data=self.ds,
            vars_to_stack=["X", "Y"],
            coords_to_stack=["coord1", "coord2"],
            pca_dim_for_rows="coord3",
        )
        pcs = self.pca.transform(
            data=self.ds,
        )
        self.assertEqual(pcs.PCs.shape[1], 5)

    def test_fit_transform(self):
        pcs = self.pca.fit_transform(
            data=self.ds,
            vars_to_stack=["X", "Y"],
            coords_to_stack=["coord1", "coord2"],
            pca_dim_for_rows="coord3",
        )
        self.assertEqual(pcs.PCs.shape[1], 5)

    def test_inverse_transform(self):
        pcs = self.pca.fit_transform(
            data=self.ds,
            vars_to_stack=["X", "Y"],
            coords_to_stack=["coord1", "coord2"],
            pca_dim_for_rows="coord3",
        )
        reconstructed_ds = self.pca.inverse_transform(PCs=pcs)
        self.assertAlmostEqual(
            self.ds.isel(coord1=5, coord2=5, coord3=5),
            reconstructed_ds.isel(coord1=5, coord2=5, coord3=5),
        )

    def test_incremental_fit(self):
        self.ipca.fit(
            data=self.ds,
            vars_to_stack=["X", "Y"],
            coords_to_stack=["coord1", "coord2"],
            pca_dim_for_rows="coord3",
        )
        self.assertEqual(self.ipca.is_fitted, True)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import pandas as pd
from typing import List, Tuple
from ._base_datamining import BaseClustering
from ..core.decorators import validate_data_mda


class MDAError(Exception):
    """
    Custom exception for MDA class.
    """

    def __init__(self, message: str = "MDA error occurred."):
        self.message = message
        super().__init__(self.message)


class MDA(BaseClustering):
    """
    Maximum Dissimilarity Algorithm (MDA) class.

    This class performs the MDA algorithm on a given dataframe.

    Attributes
    ----------
    num_centers : int
        The number of centers to use in the MDA algorithm.
    data : pd.DataFrame
        The input data.
    normalized_data : pd.DataFrame
        The normalized input data.
    data_to_fit : pd.DataFrame
        The data to fit the MDA algorithm.
    data_variables : List[str]
        A list with all data variables.
    directional_variables : List[str]
        A list with directional variables.
    fitting_variables : List[str]
        A list with fitting variables.
    custom_scale_factor : dict
        A dictionary of custom scale factors.
    scale_factor : dict
        A dictionary of scale factors (after normalizing the data).
    centroids : pd.DataFrame
        The selected centroids.
    normalized_centroids : pd.DataFrame
        The selected normalized centroids.
    centroid_iterative_indices : List[int]
        A list of iterative indices of the centroids.
    centroid_real_indices : List[int]
        The real indices of the selected centroids.

    Methods
    -------
    fit(data, directional_variables, custom_scale_factor, first_centroid_seed)
        Fit the MDA algorithm to the provided data.
    predict(data)
        Predict the nearest centroid for the provided data.
    fit_predict(data, directional_variables, custom_scale_factor, first_centroid_seed)
        Fits the MDA model to the data and predicts the nearest centroids.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from bluemath_tk.datamining.mda import MDA
    >>> data = pd.DataFrame(
    ...     {
    ...         'Hs': np.random.rand(1000) * 7,
    ...         'Tp': np.random.rand(1000) * 20,
    ...         'Dir': np.random.rand(1000) * 360
    ...     }
    ... )
    >>> mda = MDA(num_centers=10)
    >>> nearest_centroids_idxs, nearest_centroids_df = mda.fit_predict(
    ...     data=data,
    ...     directional_variables=['Dir'],
    ... )
    """

    def __init__(self, num_centers: int) -> None:
        """
        Initializes the MDA class.

        Parameters
        ----------
        num_centers : int
            The number of centers to use in the MDA algorithm.
            Must be greater than 0.

        Raises
        ------
        ValueError
            If num_centers is not greater than 0.
        """

        super().__init__()
        self.set_logger_name(name=self.__class__.__name__)
        if num_centers > 0:
            self.num_centers = int(num_centers)
        else:
            raise ValueError("Variable num_centers must be > 0")
        self._data: pd.DataFrame = pd.DataFrame()
        self._normalized_data: pd.DataFrame = pd.DataFrame()
        self._data_to_fit: pd.DataFrame = pd.DataFrame()
        self.data_variables: List[str] = []
        self.directional_variables: List[str] = []
        self.fitting_variables: List[str] = []
        self.custom_scale_factor: dict = {}
        self.scale_factor: dict = {}
        self.centroids: pd.DataFrame = pd.DataFrame()
        self.normalized_centroids: pd.DataFrame = pd.DataFrame()
        self.centroid_iterative_indices: List[int] = []
        self.centroid_real_indices: np.ndarray = np.array([])
        self.is_fitted: bool = False

    @property
    def data(self) -> pd.DataFrame:
        return self._data

    @property
    def normalized_data(self) -> pd.DataFrame:
        return self._normalized_data

    @property
    def data_to_fit(self) -> pd.DataFrame:
        return self._data_to_fit

    def _normalized_distance(
        self, array_to_compare: np.ndarray, all_rest_data: np.ndarray
    ) -> np.ndarray:
        """
        Calculate the normalized distance between the array_to_compare and all_rest_data.

        Parameters
        ----------
        array_to_compare : np.ndarray
            The array to compare against.
        all_rest_data : np.ndarray)
            The rest of the data.

        Returns
        -------
        np.ndarray
            An array of squared Euclidean distances between the two arrays for each row.

        Raises
        ------
        MDAError
            If the function is NOT called before or during fitting.

        Notes
        -----
        - IMPORTANT: Data is assumed to be normalized before calling this function.
        - The function assumes that the data_variables, directional_variables, and scale_factor
            attributes have been set.
        - The function calculates the squared sum of differences for each row.
        - DEPRECATED: directional_distance calculation.
            distance = np.absolute(array_to_compare[:, ix] - all_rest_data[:, ix])
            diff[:, ix] = np.minimum(
                distance, 1 - distance
            )
        """

        if not self.fitting_variables:
            raise MDAError(
                "_normalized_distance must be called after or during fitting, not before."
            )

        diff = np.zeros(all_rest_data.shape)

        # Calculate differences for columns
        for ix in range(len(self.fitting_variables)):
            diff[:, ix] = array_to_compare[:, ix] - all_rest_data[:, ix]
            ix = ix + 1

        # Compute the squared sum of differences for each row
        dist = np.sum(diff**2, axis=1)

        return dist

    def _nearest_indices_to_centroids(
        self, normalized_data: pd.DataFrame
    ) -> Tuple[np.ndarray, pd.DataFrame]:
        """
        Compute nearest data points to calculated centroids.

        Parameters
        ----------
        normalized_data : pd.DataFrame
            The input data to be used to compute nearest data point to centroids.

        Returns
        -------
        np.ndarray
            An array containing the index of the nearest data point to centroids.
        pd.DataFrame
            A DataFrame containing the nearest data points to centroids.

        Raises
        ------
        MDAError
            If the fit method has not been called before this method.
            Or if the data is empty.
        """

        if normalized_data.empty:
            raise MDAError("Data cannot be empty.")

        # Compute distances and store nearest distance index
        nearest_indices_array = np.zeros(self.normalized_centroids.shape[0], dtype=int)

        for i in range(self.normalized_centroids.shape[0]):
            rep = np.repeat(
                np.expand_dims(self.normalized_centroids.values[i, :], axis=0),
                normalized_data.values.shape[0],
                axis=0,
            )
            ndist = self._normalized_distance(
                array_to_compare=rep, all_rest_data=normalized_data.values
            )
            nearest_indices_array[i] = np.nanargmin(ndist)

        return nearest_indices_array, normalized_data.iloc[nearest_indices_array]

    def _nearest_indices(
        self, normalized_data: pd.DataFrame
    ) -> Tuple[np.ndarray, pd.DataFrame]:
        """
        Compute nearest centroids to the provided data.

        Parameters
        ----------
        normalized_data : pd.DataFrame
            The input data to be used to compute nearest centroids.

        Returns
        -------
        np.ndarray
            An array containing the index of the nearest centroid to the data.
        pd.DataFrame
            A DataFrame containing the nearest centroids to the data.

        Raises
        ------
        MDAError
            If the fit method has not been called before this method.
            Or if the data is empty.
        """

        if normalized_data.empty:
            raise MDAError("Data cannot be empty.")

        # Compute distances and store nearest distance index
        nearest_indices_array = np.zeros(normalized_data.shape[0], dtype=int)

        for i in range(normalized_data.shape[0]):
            rep = np.repeat(
                np.expand_dims(normalized_data.values[i, :], axis=0),
                self.normalized_centroids.values.shape[0],
                axis=0,
            )
            ndist = self._normalized_distance(
                array_to_compare=rep, all_rest_data=self.normalized_centroids.values
            )
            nearest_indices_array[i] = np.nanargmin(ndist)

        return nearest_indices_array, self.centroids.iloc[nearest_indices_array]

    @validate_data_mda
    def fit(
        self,
        data: pd.DataFrame,
        directional_variables: List[str] = [],
        custom_scale_factor: dict = {},
        first_centroid_seed: int = None,
    ) -> None:
        """
        Fit the Maximum Dissimilarity Algorithm (MDA) to the provided data.

        This method initializes centroids for the MDA algorithm using the provided
        dataframe, directional variables, and custom scale factor. It normalizes the
        data, iteratively selects centroids based on maximum dissimilarity, and
        denormalizes the centroids before returning them.

        Parameters
        ----------
        data : pd.DataFrame
            The input data to be used for the MDA algorithm.
        directional_variables : List[str], optional
            A list of names of the directional variables within the data.
            Default is [].
        custom_scale_factor : dict, optional
            A dictionary specifying custom scale factors for normalization.
            Default is {}.
        first_centroid_seed : int, optional
            The index of the first centroid to use in the MDA algorithm.
            Default is None.

        Notes
        -----
        - The function assumes that the data is validated by the `validate_data_mda`
            decorator before execution.
        - When first_centroid_seed is not provided, max value centroid is used.
        """

        self._data = data.copy()
        self.directional_variables = directional_variables.copy()
        for directional_variable in self.directional_variables:
            u_comp, v_comp = self.get_uv_components(
                x_deg=self.data[directional_variable].values
            )
            self.data[f"{directional_variable}_u"] = u_comp
            self.data[f"{directional_variable}_v"] = v_comp
        self.data_variables = list(self.data.columns)
        self.custom_scale_factor = custom_scale_factor.copy()

        # Get just the data to be used in the fitting
        self._data_to_fit = self.data.copy()
        for directional_variable in self.directional_variables:
            self.data_to_fit.drop(columns=[directional_variable], inplace=True)
        self.fitting_variables = list(self.data_to_fit.columns)

        # Normalize provided data with instantiated custom_scale_factor
        self._normalized_data, self.scale_factor = self.normalize(
            data=self.data_to_fit, custom_scale_factor=self.custom_scale_factor
        )

        # [DEPRECATED] Select the point with the maximum value in the first column of pandas dataframe
        # seed = self.normalized_data[self.normalized_data.columns[0]].idxmax()
        # Select the point with the maximum summed value
        if first_centroid_seed is not None:
            seed = first_centroid_seed
            self.logger.info(f"Using specified seed={seed} as first centroid.")
        else:
            seed = np.argmax(self.normalized_data.sum(axis=1).values)
            self.logger.info(
                f"Using max calculated value seed={seed} as first centroid."
            )

        # Initialize centroids subset
        subset = np.array(
            [self.normalized_data.values[seed]]
        )  # The row that starts as seed
        train = np.delete(self.normalized_data.values, seed, axis=0)

        # Repeat until we have the desired num_centers
        n_c = 1
        while n_c < self.num_centers:
            m2 = subset.shape[0]
            if m2 == 1:
                xx2 = np.repeat(subset, train.shape[0], axis=0)
                d_last = self._normalized_distance(
                    array_to_compare=xx2, all_rest_data=train
                )
            else:
                xx = np.array([subset[-1, :]])
                xx2 = np.repeat(xx, train.shape[0], axis=0)
                d_prev = self._normalized_distance(
                    array_to_compare=xx2, all_rest_data=train
                )
                d_last = np.minimum(d_prev, d_last)

            qerr, bmu = np.nanmax(d_last), np.nanargmax(d_last)

            if not np.isnan(qerr):
                self.centroid_iterative_indices.append(bmu)
                subset = np.append(subset, np.array([train[bmu, :]]), axis=0)
                train = np.delete(train, bmu, axis=0)
                d_last = np.delete(d_last, bmu, axis=0)

                # Log
                fmt = "0{0}d".format(len(str(self.num_centers)))
                self.logger.info(
                    "   MDA centroids: {1:{0}}/{2:{0}}".format(
                        fmt, subset.shape[0], self.num_centers
                    )
                )

            n_c = subset.shape[0]

        # De-normalize scalar and directional data
        self.normalized_centroids = pd.DataFrame(subset, columns=self.fitting_variables)
        self.centroids = self.denormalize(
            normalized_data=self.normalized_centroids, scale_factor=self.scale_factor
        )
        for directional_variable in self.directional_variables:
            self.centroids[directional_variable] = self.get_degrees_from_uv(
                xu=self.centroids[f"{directional_variable}_u"].values,
                xv=self.centroids[f"{directional_variable}_v"].values,
            )

        # Calculate the real indices of the centroids
        self.centroid_real_indices, _ = self._nearest_indices_to_centroids(
            normalized_data=self.normalized_data
        )

        # Set the fitted flag to True
        self.is_fitted = True

    def predict(self, data: pd.DataFrame) -> Tuple[np.ndarray, pd.DataFrame]:
        """
        Predict the nearest centroid for the provided data.

        Parameters
        ----------
        data : pd.DataFrame
            The input data to be used for the prediction.

        Returns
        -------
        Tuple[np.ndarray, pd.DataFrame]
            A tuple containing the nearest centroid index for each data point and the nearest centroids.
        """

        if self.is_fitted is False:
            raise MDAError("MDA model is not fitted.")
        data = data.copy()  # Avoid modifying the original data to predict
        for directional_variable in self.directional_variables:
            u_comp, v_comp = self.get_uv_components(
                x_deg=data[directional_variable].values
            )
            data[f"{directional_variable}_u"] = u_comp
            data[f"{directional_variable}_v"] = v_comp
            data.drop(columns=[directional_variable], inplace=True)
        normalized_data, _ = self.normalize(
            data=data, custom_scale_factor=self.scale_factor
        )

        return self._nearest_indices(normalized_data=normalized_data)

    def fit_predict(
        self,
        data: pd.DataFrame,
        directional_variables: List[str] = [],
        custom_scale_factor: dict = {},
        first_centroid_seed: int = None,
    ) -> Tuple[np.ndarray, pd.DataFrame]:
        """
        Fits the MDA model to the data and predicts the nearest centroids.

        Parameters
        ----------
        data : pd.DataFrame
            The input data to be used for the MDA algorithm.
        directional_variables : List[str], optional
            A list of names of the directional variables within the data.
            Default is [].
        custom_scale_factor : dict, optional
            A dictionary specifying custom scale factors for normalization.
            Default is {}.
        first_centroid_seed : int, optional
            The index of the first centroid to use in the MDA algorithm.
            Default is None.

        Returns
        -------
        Tuple[np.ndarray, pd.DataFrame]
            A tuple containing the nearest centroid index for each data point and the nearest centroids.
        """

        self.fit(
            data=data,
            directional_variables=directional_variables,
            custom_scale_factor=custom_scale_factor,
            first_centroid_seed=first_centroid_seed,
        )

        return self.predict(data=data)



import unittest
from unittest.mock import Mock
import keras


class TestResNetTraining(unittest.TestCase):
    def setUp(self):
        # Mock the data generator
        self.train_generator = Mock()
        self.train_generator.num_images = 5000
        self.train_generator.input_height = 64
        self.train_generator.input_width = 64
        self.train_generator.output_height = 64
        self.train_generator.output_width = 64
        self.train_generator.batch_size = 1

        # Define optimizer
        self.optimizer = keras.optimizers.AdamW

        # Mock the model
        self.model = Mock()
        self.model.compile = Mock()
        self.model.fit = Mock(return_value=Mock(history={"loss": [0.1, 0.05, 0.01]}))

    def test_model_training(self):
        # Compile the model
        self.model.compile(
            optimizer=self.optimizer(learning_rate=1e-4, weight_decay=1e-5),
            loss=keras.losses.mean_squared_error,
        )

        # Assert that compile was called with correct parameters
        self.model.compile.assert_called_with(
            optimizer=self.optimizer(learning_rate=1e-4, weight_decay=1e-5),
            loss=keras.losses.mean_squared_error,
        )

        # Start the train loop with the fit method
        history = self.model.fit(
            self.train_generator, initial_epoch=0, epochs=20, steps_per_epoch=500
        )

        # Assert that fit was called with correct parameters
        self.model.fit.assert_called_with(
            self.train_generator, initial_epoch=0, epochs=20, steps_per_epoch=500
        )

        # Assert that the training history is as expected
        self.assertIn("loss", history.history)
        self.assertEqual(len(history.history["loss"]), 3)
        self.assertAlmostEqual(history.history["loss"][0], 0.1)
        self.assertAlmostEqual(history.history["loss"][1], 0.05)
        self.assertAlmostEqual(history.history["loss"][2], 0.01)

        print("training complete")


if __name__ == "__main__":
    unittest.main()
